/* extension.js
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

/* exported init */

'use strict';

const { Clutter, Gio, GLib, GObject, Shell, St } = imports.gi;

const Main = imports.ui.main;
const PanelMenu = imports.ui.panelMenu;
const PopupMenu = imports.ui.popupMenu;
const SystemActions = imports.misc.systemActions;
const ExtensionUtils = imports.misc.extensionUtils;
const System = Main.panel.statusArea.aggregateMenu._system;
const SystemMenu = System.menu;

const GnomeSession = imports.misc.gnomeSession;
const SessionManager = GnomeSession.SessionManager();

const Config = imports.misc.config;
const SHELL_MAJOR_VERSION = parseInt(Config.PACKAGE_VERSION.split('.')[0]);

var DISABLE_LOCK_SCREEN_SCHEMA = "org.gnome.desktop.lockdown";

let DefaultActions;
let bindFlags = GObject.BindingFlags.DEFAULT | GObject.BindingFlags.SYNC_CREATE;
let item;
let button;

var LabelLauncher = new GObject.registerClass(
class LabelLauncher extends St.Widget {
_init() {
        this._labelText = "";
        this.label = new St.Label({ style_class: 'dash-label' });
        this.label.hide();
        Main.layoutManager.addChrome(this.label);
        this.label_actor = this.label;
        }
        
showLabel(button) {
        this.label.set_text(button.get_accessible_name());
        this.label.opacity = 0;
        this.label.show();
        
        let node = this.label.get_theme_node();
        let stageX;
        let stageY;
        let yOffset;
        
        if(SHELL_MAJOR_VERSION == 3) {
        stageX = Math.floor(button.get_transformed_position().slice()[0] + (button.get_width()/2));
        stageY = Math.floor(Main.panel.statusArea.aggregateMenu.menu.box.get_transformed_position().slice()[1] + Main.panel.statusArea.aggregateMenu.menu.box.get_height());
        yOffset = node.get_length('-x-offset');
        } else if(SHELL_MAJOR_VERSION >= 40) {
        stageX = button.get_transformed_extents().get_center().x;
        stageY = Main.panel.statusArea.aggregateMenu.menu.box.get_transformed_extents().get_bottom_right().y;  
	yOffset = node.get_length('-y-offset');
	}
        
        const labelWidth = this.label.get_width();
        const xOffset = Math.floor(labelWidth / 2);
        
        const x = stageX - xOffset;
        const y = stageY + yOffset;

        this.label.set_position(x, y);
        this.label.ease({
            opacity: 255,
            duration: 100,
            mode: Clutter.AnimationMode.EASE_OUT_QUAD,
        });
    	}
    
hideLabel() {
        this.label.ease({
            opacity: 0,
            duration: 100,
            mode: Clutter.AnimationMode.EASE_OUT_QUAD,
            onComplete: () => this.label.hide(),
        });
    	}
});

var _systemMenuButtons = new GObject.registerClass(
class SystemMenuButtons extends PanelMenu.SystemIndicator {

_init() {
    	this._showLabelTimeoutId = 0;
        this._resetHoverTimeoutId = 0;
        this._labelShowing = false;
        
	DefaultActions = new SystemActions.getDefault();
	this._settings = ExtensionUtils.getSettings();
	
	System._orientationLockItem.hide();
	System._settingsItem.hide();
	System._lockScreenItem.hide();
	SystemMenu.actor.remove_child(System._sessionSubMenu);
	
	this._createMenu();
	this._connectSettings();
	
	SystemMenu.connect('open-state-changed', (menu, open) => {
		if(!open)
		return;
		DefaultActions.forceUpdate();
	});
    	} 

_syncLabel(tooltip, button) {
        if (tooltip.child.hover) {
            if (this._showLabelTimeoutId == 0) {
            	let timeout = this._labelShowing ? 0 : 100;
                this._showLabelTimeoutId = GLib.timeout_add(GLib.PRIORITY_DEFAULT, timeout,
                    () => {
                        this._labelShowing = true;
                        tooltip.showLabel(button);
                        this._showLabelTimeoutId = 0;
                        return GLib.SOURCE_REMOVE;
                    });
                    GLib.Source.set_name_by_id(this._showLabelTimeoutId, '[gnome-shell] tooltip.showLabel');
                    if (this._resetHoverTimeoutId > 0) {
                    GLib.source_remove(this._resetHoverTimeoutId);
                    this._resetHoverTimeoutId = 0;
                }
            }
        } else {
            if (this._showLabelTimeoutId > 0)
                GLib.source_remove(this._showLabelTimeoutId);
                this._showLabelTimeoutId = 0;
            	tooltip.hideLabel();
            		if (this._labelShowing) {
                		this._resetHoverTimeoutId = GLib.timeout_add(
                    		GLib.PRIORITY_DEFAULT, 100,
                    		() => {
                        	this._labelShowing = false;
                        	this._resetHoverTimeoutId = 0;
                        	return GLib.SOURCE_REMOVE;
                    		});
                GLib.Source.set_name_by_id(this._resetHoverTimeoutId, '[gnome-shell] this._labelShowing');
            	}
            }
    	}
    	
_onDestroy() {
	if(this._resetHoverTimeoutID) {
	GLib.source_remove(this._resetHoverTimeoutID);
	this._resetHoverTimeoutID = null;
	}
	
	if (this.lockScreenSettingsChanged) {
	this.lockScreenSettings.disconnect(this.lockScreenSettingsChanged);
	}
	}

_createActionButton(iconName, accessibleName) {
        
        let roundButton = new St.Button({ reactive: true,
                                   can_focus: true,
                                   track_hover: true,
                                   accessible_name: accessibleName,
                                   x_expand: true,
                                   x_align: 2,
                                   style_class: 'rounded-menu-buttons',
                                   });
                                   
        roundButton.child = new St.Icon({ icon_name: iconName, icon_size: this._settings.get_int('icon-size') });
        roundButton.set_style("color: " + this._settings.get_string('color') + "; border-radius:" + this._settings.get_int('border-radius') + "px; padding:" + this._settings.get_int('padding') + "px;");
        
        let tooltip = new LabelLauncher();
	tooltip.child = roundButton;
	this._hookUpLabel(tooltip, roundButton);
        
        return roundButton;
    	}
    
_hookUpLabel(tooltip, button) {
        tooltip.child.connect('notify::hover', () => {
            this._syncLabel(tooltip, button);
        });
        
        tooltip.child.connect('clicked', () => {
            this._labelShowing = false;
            tooltip.hideLabel();
        });
    	}    
    
_createMenu() {
        item = new PopupMenu.PopupBaseMenuItem({ reactive: false, can_focus: false });
        let iconSize = this._settings.get_int('icon-size');
        
        // ORIENTATION BUTTON
        
        button = this._orientationButton = this._createActionButton(DefaultActions.orientation_lock_icon, DefaultActions.getName('lock-orientation'));
        button.connect('clicked', () => {
        DefaultActions.activateLockOrientation();
        });
        
        item.actor.add(button);
	DefaultActions.bind_property('can-lock-orientation', this._orientationButton, 'visible', bindFlags);
        
        DefaultActions.connect('notify::orientation-lock-icon', () => {
            let iconName = DefaultActions.orientation_lock_icon;
            let labelText = DefaultActions.getName('lock-orientation');

            button.setIcon(iconName);
            button.label.text = labelText;
        });
        
        // SETTINGS BUTTON

	let app = this._settingsApp = Shell.AppSystem.get_default().lookup_app('gnome-control-center.desktop');
        if (app) {	
        button = this._settingsButton = this._createActionButton('org.gnome.Settings-symbolic', 'Settings');
        button.connect('clicked', () => {
                        				Main.overview.hide();
                					this._settingsApp.activate();
                				      				     });
             				      				     
        item.actor.add(button);
   	} else {
            log('Missing required core component Settings, expect trouble…');
            this._settingsButton = new St.Widget();
        } 
       
        //  LOCK BUTTON
        
        button = this._lockButton = this._createActionButton('changes-prevent-symbolic', _('Lock'));
        button.connect('clicked', () => {
        DefaultActions.activateLockScreen();
        });
        
        item.actor.add(button);
        DefaultActions.bind_property('can-lock-screen', this._lockButton, 'visible', bindFlags);
        
        // SUSPEND
        
        button = this._suspendButton = this._createActionButton('media-playback-pause-symbolic', _('Suspend'));
        button.connect('clicked', () => {
        DefaultActions.activateSuspend();
        });
        
        item.actor.add(button);
        DefaultActions.bind_property('can-suspend', this._suspendButton, 'visible', bindFlags);
        
        // SWITCH USER
        
        button = this._switchUserButton = this._createActionButton('system-switch-user-symbolic', _('Switch User'));
        button.connect('clicked', () => {
        DefaultActions.activateSwitchUser();
        });
        
        item.actor.add(button);
	DefaultActions.bind_property('can-switch-user', this._switchUserButton, 'visible', bindFlags);       
        
        // LOGOUT
        
        button = this._logoutButton = this._createActionButton('system-log-out-symbolic', _('Log Out'));
        //button.connect('notify::hover', () => { this._onHover(this._logoutButton); });
        button.connect('clicked', () => {
        DefaultActions.activateLogout();
        });
        
        item.actor.add(button);
        DefaultActions.bind_property('can-logout', this._logoutButton, 'visible', bindFlags);
        
        // RESTART
        
        button = this._restartButton = this._createActionButton('system-reboot-symbolic', _('Restart'));
        button.connect('clicked', () => {
        SHELL_MAJOR_VERSION >= 40 ? DefaultActions.activateRestart() : SessionManager.RebootRemote();
        });
        
        item.actor.add(button);
	SHELL_MAJOR_VERSION >= 40 ? DefaultActions.bind_property('can-restart', this._restartButton, 'visible', bindFlags) :
					DefaultActions.bind_property('can-power-off', this._restartButton, 'visible', bindFlags);
	        
        // POWEROFF
        
        button = this._powerButton = this._createActionButton('system-shutdown-symbolic', _('Power Off'));
        button.connect('clicked', () => {
        DefaultActions.activatePowerOff();
        });
        
        item.actor.add(button);
	DefaultActions.bind_property('can-power-off', this._powerButton, 'visible', bindFlags);
	
	SystemMenu.box.add_actor(item);
	
	DefaultActions.forceUpdate();
    	}
    
_connectSettings() {
	this._settings.connect('changed::color', this._sizeChanged.bind(this));
        this._settings.connect('changed::icon-size', this._sizeChanged.bind(this));
        this._settings.connect('changed::border-radius', this._sizeChanged.bind(this));
        this._settings.connect('changed::padding', this._sizeChanged.bind(this));
        this.lockScreenSettings = new Gio.Settings({ schema: DISABLE_LOCK_SCREEN_SCHEMA });
        this.lockScreenSettingsChanged = this.lockScreenSettings.connect('changed::disable-lock-screen', this._checkLockIcon.bind(this));
    	}
    	
_checkLockIcon() {
	if(System._lockScreenItem) {
	System._lockScreenItem.hide(); }
	}
    
_sizeChanged() {
	SystemMenu.box.remove_actor(item);
	this._createMenu();
        }
});

function init() {
}

let modifiedMenu;

function enable() {
modifiedMenu = new _systemMenuButtons();
}

function disable() {
	SystemMenu.box.remove_actor(item);
	if(Main.sessionMode.currentMode !== 'unlock-dialog') {
		System._orientationLockItem.show();
		DefaultActions.bind_property('can-lock-orientation', System._orientationLockItem, 'visible', bindFlags);
		System._settingsItem.show();
		System._lockScreenItem.show();
	}
	SystemMenu.box.insert_child_at_index(System._orientationLockItem, SystemMenu.numMenuItems);
	SystemMenu.box.insert_child_at_index(System._sessionSubMenu, SystemMenu.numMenuItems);
	modifiedMenu._onDestroy();
	modifiedMenu = null;
}
